#!/usr/bin/python
"""
The module provides a client to the PackageKit DBus interface. It allows to
perform basic package manipulation tasks in a cross distribution way, e.g.
to search for packages, install packages or codecs.
"""
# Copyright (C) 2008 Canonical Ltd.
# Copyright (C) 2008 Aidan Skinner <aidan@skinner.me.uk>
# Copyright (C) 2008 Martin Pitt <martin.pitt@ubuntu.com>
# Copyright (C) 2008 Tim Lauridsen <timlau@fedoraproject.org>
# Copyright (C) 2008-2009 Sebastian Heinlein <devel@glatzor.de>
#
# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import locale
import weakref

import dbus
import dbus.glib
import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
import gobject

import enums
import debconf


class AptDaemonError(Exception):
    """Wrapper for a ErrorCode message"""
    def __init__(self, code, details=None):
        Exception.__init__(self)
        self.code = code
        self.details = details

    def __str__(self):
        return "%s: %s" % (self.code, self.details)


class MemoizedTransaction(type):

    """Metaclass to cache transactions."""

    cache = weakref.WeakValueDictionary()

    def __call__(mcs, *args):
        tid = args[0]
        try:
            return mcs.cache[tid]
        except KeyError:
            mcs.cache[tid] = value = \
                    super(MemoizedTransaction, mcs).__call__(*args)
            return value


class MemoizedMixIn(MemoizedTransaction, gobject.GObjectMeta):

    """Helper meta class for merging"""


class AptTransaction(gobject.GObject):

    """Represents an aptdaemon transaction. It allows asynchronous and
    synchronous processing.
    """

    __metaclass__ = MemoizedMixIn

    __gsignals__ = {"finished": (gobject.SIGNAL_RUN_FIRST,
                                 gobject.TYPE_NONE,
                                 (gobject.TYPE_INT,)),
                    "error": (gobject.SIGNAL_RUN_FIRST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_INT, gobject.TYPE_STRING)),
                    "role-changed": (gobject.SIGNAL_RUN_FIRST,
                                     gobject.TYPE_NONE,
                                     (gobject.TYPE_INT,)),
                    "terminal-attachable-changed": (gobject.SIGNAL_RUN_FIRST,
                                                    gobject.TYPE_NONE,
                                                    (gobject.TYPE_BOOLEAN,)),
                    "cancellable-changed": (gobject.SIGNAL_RUN_FIRST,
                                            gobject.TYPE_NONE,
                                            (gobject.TYPE_BOOLEAN,)),
                    "status-changed": (gobject.SIGNAL_RUN_FIRST,
                                       gobject.TYPE_NONE,
                                       (gobject.TYPE_INT,)),
                    "status-details-changed": (gobject.SIGNAL_RUN_FIRST,
                                               gobject.TYPE_NONE,
                                               (gobject.TYPE_STRING,)),
                    "progress-changed": (gobject.SIGNAL_RUN_FIRST,
                                 gobject.TYPE_NONE,
                                 (gobject.TYPE_INT,)),
                    "progress-details-changed": (gobject.SIGNAL_RUN_FIRST,
                                         gobject.TYPE_NONE,
                                         (gobject.TYPE_INT, gobject.TYPE_INT,
                                          gobject.TYPE_INT, gobject.TYPE_INT,
                                          gobject.TYPE_INT, gobject.TYPE_INT)),
                    "paused": (gobject.SIGNAL_RUN_FIRST,
                               gobject.TYPE_NONE,
                               ()),
                    "resumed": (gobject.SIGNAL_RUN_FIRST,
                                gobject.TYPE_NONE,
                                ()),
                    "allow-unauthenticated-changed": (gobject.SIGNAL_RUN_FIRST,
                                                      gobject.TYPE_NONE,
                                                      (gobject.TYPE_BOOLEAN,)),
                    "locale-changed": (gobject.SIGNAL_RUN_FIRST,
                                       gobject.TYPE_NONE,
                                       (gobject.TYPE_STRING,)),
                    "terminal-changed": (gobject.SIGNAL_RUN_FIRST,
                                         gobject.TYPE_NONE,
                                         (gobject.TYPE_STRING,)),
                    "debconf-socket-changed": (gobject.SIGNAL_RUN_FIRST,
                                               gobject.TYPE_NONE,
                                               (gobject.TYPE_STRING,)),
                    "medium-required": (gobject.SIGNAL_RUN_FIRST,
                                        gobject.TYPE_NONE,
                                        (gobject.TYPE_STRING,
                                         gobject.TYPE_STRING)),
                    "config-file-prompt": (gobject.SIGNAL_RUN_FIRST,
                                           gobject.TYPE_NONE,
                                           (gobject.TYPE_STRING,
                                            gobject.TYPE_STRING)),
                    }

    def __init__(self, tid, bus=None):
        gobject.GObject.__init__(self)
        self.tid = tid
        self.role = enums.ROLE_UNSET
        self.error_code = None
        self.error_details = None
        self.exit = None
        self.cancellable = False
        self.term_attachable = False
        self.required_medium = None
        self.config_file_prompt = None
        self.status = None
        self.status_details = ""
        self.progress = 0
        self.paused = False
        self._locale = ""
        self._exit_handler = None
        self._method = None
        self._args = []
        self._debconf_helper = None
        # Connect the signal handlers to the DBus iface
        if not bus:
            bus = dbus.SystemBus()
        self._proxy = bus.get_object("org.debian.apt", tid)
        self._iface = dbus.Interface(self._proxy, "org.debian.apt.transaction")
        # Watch for a crashed daemon which orphaned the dbus object
        self._owner_watcher = bus.watch_name_owner("org.debian.apt",
                                                   self._on_name_owner_changed)
        # main signals
        self._signal_matcher = \
            self._iface.connect_to_signal("PropertyChanged",
                                          self._on_property_changed,
                                          utf8_strings=True)

    def _on_name_owner_changed(self, connection):
        """Fail the transaction if the daemon died."""
        if connection == "" and not self.exit:
            self._on_property_changed("Error", (enums.ERROR_DAEMON_DIED,
                                                "It seems that the daemon "
                                                "died."))
            self._on_property_changed("Cancellable", False)
            self._on_property_changed("TerminalAttachable", False)
            self._on_property_changed("ExitState", enums.EXIT_FAILED)

    def _on_property_changed(self, property_name, value):
        """Callback for the PropertyChanged signal."""
        if property_name == "TerminalAttachable":
            self.term_attachable = value
            self.emit("terminal-attachable-changed", value)
        elif property_name == "Cancellable":
            self.cancellable = value
            self.emit("cancellable-changed", value)
        elif property_name == "DebconfSocket":
            self.emit("debconf-socket-changed", value)
        elif property_name == "AllowUnauthenticated":
            self.emit("allow-unauthenticated-changed", value)
        elif property_name == "Terminal":
            self.emit("terminal-changed", value)
        elif property_name == "Locale":
            self.locale = value
            self.emit("locale-changed", value)
        elif property_name == "Role":
            self.role = value
            self.emit("role-changed", value)
        elif property_name == "Status":
            self.status = value
            self.emit("status-changed", value)
        elif property_name == "StatusDetails":
            self.status_details = value
            self.emit("status-details-changed", value)
        elif property_name == "Progress":
            self.progress = value
            self.emit("progress-changed", value)
        elif property_name == "ConfigFilePrompt":
            self.config_file_prompt = value
            self.emit("config-file-prompt", *value)
        elif property_name == "Paused":
            self.paused = value
            if value:
                self.emit("paused")
            else:
                self.emit("resumed")
        elif property_name == "RequiredMedium":
            self.required_medium = value
            self.emit("medium-required", *value)
        elif property_name == "ProgressDetailsChanged":
            self.progress_details = value
            self.emit("progress-details-changed", *value)
        elif property_name == "Error":
            self.error_code, self.error_details = value
            self.error = AptDaemonError(self.error_code, self.error_details)
            self.emit("error", *value)
        elif property_name == "ExitState":
            self.exit = value
            self.emit("finished", value)
            self._owner_watcher.cancel()
            if self._debconf_helper:
                self._debconf_helper.stop()
            self.detach()

    def _set_method(self, method, *args):
        """Setup the method of the DBus interface which should be handled"""
        self._method = self._iface.get_dbus_method(method)
        self._args = args

    def run(self, error_handler=None, reply_handler=None):
        """Start processing the transaction.

        Keyword arguemnts:
        error_handler -- callback whill will called in the case of an error
        reply_handler -- callback which will called if the transaction is done
        """
        try:
            return self._method(timeout=250,
                                error_handler=error_handler,
                                reply_handler=reply_handler, *self._args)
        except Exception, error:
            if error_handler:
                error_handler(error)
            else:
                raise

    def cancel(self, reply_handler=None, error_handler=None):
        """Cancel the running transaction."""
        self._iface.Cancel(reply_handler=reply_handler,
                           error_handler=error_handler)

    def set_allow_unauthenticated(self, allow_unauthenticated, 
                            reply_handler=None, error_handler=None):
        """Set the allow_unauthenticated property"""
        if reply_handler:
            _reply_handler = lambda x: reply_handler(self)
        else:
            _reply_handler = None
        self._proxy.Set("org.debian.apt.transaction", 
                        "AllowUnauthenticated", allow_unauthenticated,
                        dbus_interface=dbus.PROPERTIES_IFACE,
                        reply_handler=_reply_handler,
                        error_handler=error_handler)

    def set_debconf_frontend(self, frontend, reply_handler=None,
                             error_handler=None):
        """Set the debconf socket """
        if reply_handler:
            _reply_handler = lambda x: reply_handler(self)
        else:
            _reply_handler = None
        self._debconf_helper = debconf.DebconfProxy(frontend)
        self._proxy.Set("org.debian.apt.transaction", "DebconfSocket",
                        self._debconf_helper.socket_path,
                        dbus_interface=dbus.PROPERTIES_IFACE,
                        reply_handler=_reply_handler,
                        error_handler=error_handler)
        self._debconf_helper.start()

    def set_terminal(self, ttyname, reply_handler=None, error_handler=None):
        """Set the controlling terminal."""
        if reply_handler:
            _reply_handler = lambda: reply_handler(self)
        else:
            _reply_handler = None
        self._proxy.Set("org.debian.apt.transaction", "Terminal", ttyname,
                        dbus_interface=dbus.PROPERTIES_IFACE,
                        reply_handler=_reply_handler,
                        error_handler=error_handler)

    def attach(self, reply_handler=None, error_handler=None):
        """Request the current state from the transaction."""
        self._iface.Attach(reply_handler=reply_handler,
                           error_handler=error_handler)

    def detach(self):
        """Stop monitoring the progress of the transaction."""
        self._signal_matcher.remove()
        del self._signal_matcher

    def set_locale(self, locale_name, reply_handler=None, error_handler=None):
        """Set the language."""
        if reply_handler:
            _reply_handler = lambda: reply_handler(self)
        else:
            _reply_handler = None
        self._proxy.Set("org.debian.apt.transaction", "Locale", locale_name,
                        dbus_interface=dbus.PROPERTIES_IFACE,
                        reply_handler=_reply_handler,
                        error_handler=error_handler)

    def provide_medium(self, medium, reply_handler=None, error_handler=None):
        """Continue a paused transaction which waited for a medium to install
        from.
        """
        self._iface.ProvideMedium(medium, reply_handler=reply_handler,
                                  error_handler=error_handler)

    def config_file_prompt_answer(self, config, answer, reply_handler=None, 
                                  error_handler=None):
        """Continue a paused transaction that waits for config file prompt
        """
        self._iface.ConfigFilePromptAnswer(config, answer, 
                                           reply_handler=reply_handler,
                                           error_handler=error_handler)

class AptClient:
    """Aptdaemon client wrapper class"""
    def __init__(self):
        """
        Initialize the client.
        """
        self.apt_control = None
        self.bus = dbus.SystemBus()
        self._locale = "%s.%s" % locale.getdefaultlocale()
        self.terminal = None

    def get_trusted_vendor_keys(self, reply_handler=None, error_handler=None):
        """Return the list of trusted vendors."""
        daemon = get_aptdaemon()
        keys = daemon.GetTrustedVendorKeys(reply_handler=reply_handler,
                                           error_handler=error_handler)
        return keys

    def upgrade_system(self, safe_mode=True, wait=False, reply_handler=None,
                       error_handler=None):
        """Upgrade system."""
        return self._run_transaction("UpgradeSystem", [safe_mode],
                                     wait, reply_handler, error_handler)

    def install_packages(self, package_names, wait=False, reply_handler=None,
                         error_handler=None):
        """Install packages by name."""
        return self._run_transaction("InstallPackages", [package_names],
                                     wait, reply_handler, error_handler)

    def add_repository(self, src_type, uri, dist, comps=[], comment="", 
                       sourcesfile="", reply_handler=None, error_handler=None):
        """Add repository to the sources list.

        Keyword arguments:
        src_type -- the type of the entry (deb, deb-src)
        uri -- the main repository uri (e.g. http://archive.ubuntu.com/ubuntu)
        dist -- the distribution to use (e.g. karmic, "/")
        comps -- a (possible empty) list of components (main, restricted)
        comment -- an (optional) comment
        sourcesfile -- an (optinal) filename in sources.list.d 
        """
        daemon = get_aptdaemon()
        # dbus can not deal with empty lists and will error
        if comps == []:
            comps = [""]
        return daemon.AddRepository(src_type, uri, dist, comps, comment,
                                    sourcesfile, reply_handler=reply_handler,
                                    error_handler=error_handler)

    def add_vendor_key_from_file(self, path, wait=False, reply_handler=None,
                                 error_handler=None):
        """Add signing key from the given file to the trusted vendors."""
        return self._run_transaction("AddVendorKeyFromFile", [path],
                                     wait, reply_handler, error_handler)

    def remove_vendor_key(self, fingerprint, wait=False, reply_handler=None,
                          error_handler=None):
        """Remove key from the list of trusted vendors.

        Keyword arguments:
        fingerprint -- the id of the key which should be removed
        """
        return self._run_transaction("RemoveVendorKey", [fingerprint],
                                     wait, reply_handler, error_handler)

    def install_file(self, path, wait=False, reply_handler=None,
                     error_handler=None):
        """Install local package file."""
        return self._run_transaction("InstallFile", [path],
                                     wait, reply_handler, error_handler)

    def upgrade_packages(self, package_names, wait=False, reply_handler=None,
                         error_handler=None):
        """Upgrade packages by name."""
        return self._run_transaction("UpgradePackages", [package_names],
                                     wait, reply_handler, error_handler)

    def remove_packages(self, package_names, wait=False, reply_handler=None,
                        error_handler=None):
        """Remove packages by name."""
        return self._run_transaction("RemovePackages", [package_names],
                                     wait, reply_handler, error_handler)

    def commit_packages(self, install, reinstall, remove, purge, upgrade,
                        wait=False, reply_handler=None, error_handler=None):
        """Perform complex packages changes."""
        return self._run_transaction("CommitPackages",
                                     [install, reinstall, remove, purge,
                                      upgrade],
                                     wait, reply_handler, error_handler)

    def update_cache(self, wait=False, reply_handler=None, error_handler=None):
        """Update repository information."""
        return self._run_transaction("UpdateCache", [], wait, reply_handler,
                                     error_handler)

    def _run_transaction(self, method_name, args, wait, reply_handler,
                         error_handler):
        """Run the given method in a new transaction."""
        def reply(trans):
            """Handle the return of the Transaction."""
            if wait:
                #FIXME: not implemented
                if error_handler:
                    error_handler(NotImplemented)
                else:
                    raise NotImplemented
            else:
                if async:
                    reply_handler(trans)
                else:
                    return trans

        def set_terminal(trans):
            """Setup the attachable terminal for the transaction."""
            if self.terminal:
                if async:
                    trans.set_terminal(self.terminal,
                                       reply_handler=reply,
                                       error_handler=error_handler)
                    return
                else:
                    trans.set_terminal(self.terminal)
            return reply(trans)

        def set_locale(tid):
            """Setup the language of the transaction."""
            trans = AptTransaction(tid, self.bus)
            trans._set_method(method_name, *args)
            if self._locale is not None:
                if async:
                    trans.set_locale(self._locale,
                                     reply_handler=set_terminal,
                                     error_handler=error_handler)
                    return
                else:
                    trans.set_locale(self._locale)
            return set_terminal(trans)

        def on_daemon_failed(error):
            """Error callback which tries to catch a timeout of the daemon."""
            if hasattr(error, "get_dbus_name") and \
               error.get_dbus_name() == "org.freedesktop.DBus.Error." \
                                        "ServiceUnknown":
                # Catch if the daemon has timed out
                self.apt_control = get_aptdaemon()
                self.apt_control.RequestTransaction(
                                    reply_handler=set_locale,
                                    error_handler=error_handler)

        async = reply_handler and error_handler
        # Request a new transaction
        if not self.apt_control:
            self.apt_control = get_aptdaemon()
        if async:
            self.apt_control.RequestTransaction(
                                    reply_handler=set_locale,
                                    error_handler=on_daemon_failed)
        else:
            try:
                tid = self.apt_control.RequestTransaction()
            except dbus.DBusException, error:
                if error._dbus_error_name == "org.freedesktop.DBus.Error." \
                                             "ServiceUnknown":
                    # Catch if the daemon has timed out
                    self.apt_control = get_aptdaemon()
                    tid = self.apt_control.RequestTransaction()
            return set_locale(tid)


def get_transaction(tid, bus=None):
    """Return an AptTransaction instance connected to the given transaction.

    Keyword arguments:
    tid -- the identifier of the transaction
    bus -- the D-Bus connection that should be used (defaults to the system bus)
    """
    if not bus:
        bus = dbus.SystemBus()
    trans = AptTransaction(tid, bus)
    trans.attach()
    return trans

def get_aptdaemon(bus=None):
    """Return the aptdaemon D-Bus interface.

    Keyword argument:
    bus -- the D-Bus connection that should be used (defaults to the system bus)
    """
    if not bus:
        bus = dbus.SystemBus()
    return dbus.Interface(bus.get_object("org.debian.apt",
                                         "/org/debian/apt",
                                         False),
                          "org.debian.apt")

# vim:ts=4:sw=4:et
